
#ifndef ___egg_marshal_MARSHAL_H__
#define ___egg_marshal_MARSHAL_H__

#include	<glib-object.h>

G_BEGIN_DECLS

/* VOID:INT,INT (./eggmarshalers.list:1) */
G_GNUC_INTERNAL void _egg_marshal_VOID__INT_INT (GClosure     *closure,
                                                 GValue       *return_value,
                                                 guint         n_param_values,
                                                 const GValue *param_values,
                                                 gpointer      invocation_hint,
                                                 gpointer      marshal_data);

G_END_DECLS

#endif /* ___egg_marshal_MARSHAL_H__ */

